// created: 1/3/21, Yvan Frey, Liquid Thought, LLC
// title: google interview airline algorithm problem
// swift version: 5.5

// figure least amount of routes to add to fly from starting airport to all airports

let airports = ["BGI", "CDG", "DEL", "DOH", "DSM", "EWR", "EYW", "HND", "ICN", "JFK",
                "LGA", "LHR", "ORD", "SAN", "SFO", "SIN", "TLV", "BUD"] // 18

let routes = [
["DSM", "ORD"],
["ORD", "BGI"],
["BGI", "LGA"],
["SIN", "CDG"],
["CDG", "SIN"],
["CDG", "BUD"],
["DEL", "DOH"],
["DEL", "CDG"],
["TLV", "DEL"],
["EWR", "HND"],
["HND", "ICN"],
["HND", "JFK"],
["ICN", "JFK"],
["JFK", "LGA"],
["EYW", "LHR"],
["LHR", "SFO"],
["SFO", "SAN"],
["SFO", "DSM"],
["SAN", "EYW"]
]

let startingAirport = "LGA"

// go backwards from the route's tos
func buildAirportsReachableFrom(_ airports: [String], _ routes: [[String]], _ startingAirport: String) -> [String: [String]] {

    // nested recursion function
    func airportReachableFrom(_ airportsReachableFromKey: String, _ airport: String, _ routes: [[String]], _ airportsReachableFrom: inout [String]) {
        for route in routes {
            if route[1] == airport {
                if !airportsReachableFrom.contains(route[0]) && route[0] != airportsReachableFromKey {
                    airportsReachableFrom.append(route[0])
                    airportReachableFrom(airportsReachableFromKey, route[0], routes, &airportsReachableFrom)
                }
            }
        }
        return
    }

    var airportsReachableFrom = [String: [String]]()
    for airport in airports {
        if airport != startingAirport { // don't bother looking at routes to the starting airport
            airportsReachableFrom[airport] = [String]()
            // recurse
            airportReachableFrom(airport, airport, routes, &airportsReachableFrom[airport]!)
        }
    }
    return airportsReachableFrom
}

// invert the list
func buildAirportsServingRoutesTo(_ airportsReachableFrom: [String: [String]]) -> [String: [String]] {
    var airportsTo = [String: [String]]()
    for airportFrom in airportsReachableFrom {
        for airport in airportFrom.value {
            if airportsTo[airport] == nil {
                airportsTo[airport] = [String]()
            }
            var airports = airportsTo[airport]!
            airports.append(airportFrom.key)
            airportsTo[airport] = airports
        }
    }
    return airportsTo
}

// order the list by biggest hub first
func orderAirportsServingRoutesTo(_ airportsServingRoutesTo: [String: [String]]) -> [(String, [String])] {
    
    // nested recursion function
    func orderAirports(_ airportsByNumberOfTos: inout [(String, [String])], _ airportsServingRoutesTo: inout [String: [String]]) {
        // get airport that serves most routes
        var airportWithMostRoutes = ("", [String]()) // lame initialization
        for airport in airportsServingRoutesTo {
            if airport.value.count > airportWithMostRoutes.1.count {
                airportWithMostRoutes = (airport.key, airport.value)
            }
        }
        
        if airportWithMostRoutes.1.count == 0 {
            return
        }
        
        airportsByNumberOfTos.append(airportWithMostRoutes)

        // remove tos that are already covered in airport with more routes
        for airportServingRoutesTo in airportsServingRoutesTo {
            if airportServingRoutesTo.key != airportWithMostRoutes.0 {

                // append the airport in the key
                var airportWithMostRoutesInclKey = airportWithMostRoutes.1
                airportWithMostRoutesInclKey.append(airportWithMostRoutes.0)

                airportsServingRoutesTo[airportServingRoutesTo.key] = Array(Set(airportServingRoutesTo.value).subtracting(airportWithMostRoutesInclKey))
            }
        }
        airportsServingRoutesTo.removeValue(forKey: airportWithMostRoutes.0)

        if airportsServingRoutesTo.count > 0 {
            orderAirports(&airportsByNumberOfTos, &airportsServingRoutesTo)
        }
    }
    
    var airportsByNumberOfTos = [(String, [String])]() // array of tuples
    var airportsServingRoutesToCopy = airportsServingRoutesTo
    
    orderAirports(&airportsByNumberOfTos, &airportsServingRoutesToCopy)
    return airportsByNumberOfTos
}

func determineRoutesToAdd(_ startingAirport: String, _ airports: [String], _ orderedAirportsServingRoutesTo: [(String, [String])]) -> [[String]] {
    var routesToAdd = [[String]]()
    var airportsCopy = airports

    // remove starting airport
    airportsCopy = airportsCopy.filter {$0 != startingAirport}

    // add routes to all airports that have tos
    for orderedAirportsServingRouteTo in orderedAirportsServingRoutesTo {
        routesToAdd.append([startingAirport, orderedAirportsServingRouteTo.0])

        // remove the airport that a route was added to from the copy
        airportsCopy = airportsCopy.filter {$0 != orderedAirportsServingRouteTo.0}
        
        // if all airports are covered
        if airportsCopy.isEmpty {
            break
        }

        // remove airports that have routes to them from the copy
        for airportServed in orderedAirportsServingRouteTo.1 {
            airportsCopy = airportsCopy.filter {$0 != airportServed}
        }
    }
    
    // add direct routes to the tos that have no routes to them
    for airport in airportsCopy {
        routesToAdd.append([startingAirport, airport])
    }
    return routesToAdd
}

let airportsReachableFrom = buildAirportsReachableFrom(airports, routes, startingAirport)
var airportsServingRoutesTo = buildAirportsServingRoutesTo(airportsReachableFrom)
let orderedAirportsServingRoutesTo = orderAirportsServingRoutesTo(airportsServingRoutesTo)
let routesToAdd = determineRoutesToAdd(startingAirport, airports, orderedAirportsServingRoutesTo)
print("minimum routes to add: \(routesToAdd.count)")
print("routes:\n\(routesToAdd)")
